#undef IGL_STATIC_LIBRARY
#include <igl/readOBJ.h>
#include <igl/readDMAT.h>
#include <igl/viewer/Viewer.h>
#include <igl/barycenter.h>
#include <igl/avg_edge_length.h>
#include <vector>
#include <igl/n_polyvector.h>
#include <igl/angle_bound_frame_fields.h>
#include <stdlib.h>
#include <igl/jet.h>

// Input mesh
Eigen::MatrixXd V;
Eigen::MatrixXi F;

// Face barycenters
Eigen::MatrixXd B;


// Scale for visualizing the fields
double global_scale;

// Input constraints
Eigen::VectorXi isConstrained;
Eigen::MatrixXd constraints;

Eigen::MatrixXd smooth_pvf;
Eigen::MatrixXd angle_bound_pvf;
igl::AngleBoundFFSolverData<Eigen::MatrixXd, Eigen::MatrixXi> *csdata;

int maxIter = 100;
int iterDone = 0;
// double lambdaInit = 100;
// double lambdaMultFactor = 1.5;
double lambdaInit = .1;
double lambdaMultFactor = 1.01;
bool doHardConstraints = false;

int mode;

void computeAngles(const Eigen::MatrixXd &ff, Eigen::VectorXd &angles)
{
 angles.resize(ff.rows(),1);
  // int num =0;
 for (int i =0; i<ff.rows(); ++i)
 {
  Eigen::RowVector3d u = (ff.block(i,0,1,3)); u.normalize();
  Eigen::RowVector3d v = (ff.block(i,3,1,3)); v.normalize();
  double s = (u.cross(v)).norm();
  double c = fabs(u.dot(v));
  angles[i] = atan2(s,c);
  //  num += (angles[i]<70*M_PI/180);
  }
  // std::cerr<<"out of bound:"<<num<<std::endl;
}

void getAngleColor(const Eigen::MatrixXd &ff, Eigen::MatrixXd &C)
{
  Eigen::VectorXd angles;
  computeAngles(ff, angles);
  Eigen::VectorXd val = 0.5*M_PI*Eigen::VectorXd::Ones(angles.rows(),1)-angles;
  igl::jet(val, 0, 20*M_PI/180., C);
}

bool pre_draw(igl::Viewer & viewer)
{
  using namespace Eigen;
  using namespace std;
  double lambdaOut;

  if (mode == 1)
  {
    if (iterDone <maxIter)
    {
      igl::angle_bound_frame_fields(*csdata,
                                     70,
                                     isConstrained,
                                     angle_bound_pvf,
                                     angle_bound_pvf,
                                     1,
                                     lambdaInit,
                                     lambdaMultFactor,
                                     doHardConstraints,
                                     &lambdaOut);
      iterDone ++;
      // lambdaInit = lambdaOut;
      lambdaInit *=1.1;

    }
    viewer.data.lines.resize(0,9);
    viewer.data.add_edges(B - global_scale*angle_bound_pvf.block(0,0,F.rows(),3) ,
                      B + global_scale*angle_bound_pvf.block(0,0,F.rows(),3),
                      Eigen::RowVector3d(0,0,1));
    viewer.data.add_edges(B - global_scale*angle_bound_pvf.block(0,3,F.rows(),3),
                      B + global_scale*angle_bound_pvf.block(0,3,F.rows(),3),
                      Eigen::RowVector3d(0,0,1));
  }
  return false;
}


bool key_down(igl::Viewer& viewer, unsigned char key, int modifier)
{
  using namespace std;
  using namespace Eigen;

  // Highlight in red the constrained faces
  MatrixXd CC = MatrixXd::Constant(F.rows(),3,1);
  for (unsigned i=0; i<F.rows();++i)
    if (isConstrained[i])
      CC.row(i) << 1, 0, 0;

  if (key <'1' || key >'5')
  {
    return false;
  }
  mode = 0;
  if (key == '1')
  {
    viewer.data.lines.resize(0,9);
    viewer.data.set_colors(CC);
    // Frame field constraints
    MatrixXd F1_t = MatrixXd::Zero(F.rows(),3);
    MatrixXd F2_t = MatrixXd::Zero(F.rows(),3);
    for (unsigned i=0; i<F.rows();++i)
      if (isConstrained[i])
      {
        F1_t.row(i) = constraints.block(i,0,1,3);
        F2_t.row(i) = constraints.block(i,3,1,3);
      }
    viewer.data.add_edges (B - global_scale*F1_t, B + global_scale*F1_t , Eigen::RowVector3d(0,0,1));
    viewer.data.add_edges (B - global_scale*F2_t, B + global_scale*F2_t , Eigen::RowVector3d(0,0,1));
  }
  if (key == '2')
  {
    viewer.data.lines.resize(0,9);
    viewer.data.add_edges (B - global_scale*smooth_pvf.block(0,0,F.rows(),3),
                      B + global_scale*smooth_pvf.block(0,0,F.rows(),3),
                      Eigen::RowVector3d(0,1,0));
    viewer.data.add_edges (B - global_scale*smooth_pvf.block(0,3,F.rows(),3),
                      B + global_scale*smooth_pvf.block(0,3,F.rows(),3),
                      Eigen::RowVector3d(0,1,0));
    viewer.data.set_colors(CC);
  }
  if (key == '3')
  {
    viewer.data.lines.resize(0,9);
    MatrixXd C;
    getAngleColor(smooth_pvf, C);
    viewer.data.set_colors(C);
  }

  if (key == '4')
  {
    // viewer.data.add_edges (B - global_scale*angle_bound_pvf.block(0,0,F.rows(),3),
                      // B + global_scale*angle_bound_pvf.block(0,0,F.rows(),3),
                      // Eigen::RowVector3d(0,1,0));
    // viewer.data.add_edges (B - global_scale*angle_bound_pvf.block(0,3,F.rows(),3),
                      // B + global_scale*angle_bound_pvf.block(0,3,F.rows(),3),
                      // Eigen::RowVector3d(0,1,0));
    viewer.data.lines.resize(0,9);
    viewer.data.set_colors(CC);
    mode = 1;
  }
  if (key == '5')
  {
    viewer.data.lines.resize(0,9);
    MatrixXd C;
    getAngleColor(angle_bound_pvf, C);
    viewer.data.set_colors(C);
  }

  return false;
}

int main(int argc, char *argv[])
{
  using namespace Eigen;
  using namespace std;
  // Load a mesh in OBJ format
  igl::readOBJ("../data/teddy.obj", V, F);

  // Compute face barycenters
  igl::barycenter(V, F, B);

  // Compute scale for visualizing fields
  global_scale =  .5*igl::avg_edge_length(V, F);

  // Load constraints
  MatrixXd temp;
  igl::readDMAT("../data/teddy.dmat",temp);
  isConstrained = temp.block(0,0,temp.rows(),1).cast<int>();
  constraints = temp.block(0,1,temp.rows(),temp.cols()-1);

  int numConstrained = isConstrained.sum();
  VectorXi b(numConstrained);
  MatrixXd bc(numConstrained,constraints.cols());
  int ind = 0;
  for (int fi = 0; fi <F.rows(); ++fi)
    if (isConstrained[fi])
      {
        b[ind] = fi;
        bc.row(ind) = constraints.row(fi);
        ind ++;
      }

  // Interpolated PolyVector field
  Eigen::MatrixXd pvf;
  igl::n_polyvector(V, F, b, bc, smooth_pvf);

  // Initialize conjugate field with smooth field
  csdata = new igl::AngleBoundFFSolverData<Eigen::MatrixXd,Eigen::MatrixXi>(V,F);
  angle_bound_pvf = smooth_pvf;


  igl::Viewer viewer;


  // Launch the viewer
  viewer.data.clear();
  viewer.data.set_mesh(V, F);
  viewer.core.show_lines = false;
  viewer.core.show_texture = false;
  viewer.callback_pre_draw = &pre_draw;
  viewer.callback_key_down = &key_down;
  viewer.core.is_animating = true;
  viewer.core.animation_max_fps = 30.;

  // Plot the original mesh with a texture parametrization
  key_down(viewer,'1',0);

  viewer.launch();
}
